#include "byte.h"
#include "stralloc.h"
#include "uint_t.h"
#include "dns.h"
#include "logmsg.h"

// stralloc out = {0};
static char *q = 0;

int dns_tlsa_packet(stralloc *out,const char *buf,unsigned int len)
{
  unsigned int pos;
  char header[12];
  uint16 datalen;
  uint16 numanswers;
  int ranswers = 0;

  if (!stralloc_copys(out,"")) return DNS_MEM;

  pos = dns_packet_copy(buf,len,0,header,12); if (!pos) return DNS_ERR;
  uint16_unpack_big(header + 6,&numanswers);
  pos = dns_packet_skipname(buf,len,pos); if (!pos) return DNS_ERR;
  pos += 4;

  while (numanswers--) {
    pos = dns_packet_skipname(buf,len,pos); if (!pos) return DNS_ERR;
    pos = dns_packet_copy(buf,len,pos,header,10); if (!pos) return DNS_ERR;
    uint16_unpack_big(header + 8,&datalen);
    if (byte_equal(header,2,DNS_T_TLSA))
      if (byte_equal(header + 2,2,DNS_C_IN)) {
        if (pos + datalen > len) return DNS_ERR;
        if (!stralloc_catb(out,buf + pos,datalen)) return DNS_MEM;
      }
    pos += datalen;
    ++ranswers;
  }
  if (!stralloc_0(out)) return DNS_MEM;

  return ranswers;
}

int dns_tlsa(stralloc *out,const stralloc *fqdn)
{
  int rc = 0;

  if (dns_domain_fromdot(&q,fqdn->s,fqdn->len) <= 0) return DNS_ERR;
  if (dns_resolve(q,DNS_T_TLSA) >= 0) {
    if ((rc = dns_tlsa_packet(out,dns_resolve_tx.packet,dns_resolve_tx.packetlen)) < 0) return DNS_ERR;
    dns_transmit_free(&dns_resolve_tx);
    dns_domain_free(&q);
  }

  return rc;
}
